#######################################################################################
### Abaqus script to create a model of a bending 				    ###
### fiber-reinforced pneumatic soft actuator 					    ###
### Fionnuala Connolly								    ###
### May 2016									    ###
### This script is compatible with Abaqus v6.14					    ###
### To modify actuator geometry or material properties, see PARAMETERS_BENDING.py ###
#######################################################################################
from part import *
from material import *
from section import *
from assembly import *
from step import *
from interaction import *
from load import *
from mesh import *
from optimization import *
from job import *
from sketch import *
from visualization import *
from connectorBehavior import *
import math
import numpy as np
import os
import datetime
import shutil
from odbAccess import *
from array import *
import time
import random
import bisect
import regionToolset
Mdb()

############################################################
#LOAD PARAMETERS
############################################################
execfile('PARAMETERS_BENDING.py')			

############################################################
# CREATE CYLINDRICAL TUBE 					 
############################################################

mdb.models['Model-1'].ConstrainedSketch(name='__profile__', sheetSize=200.0)
mdb.models['Model-1'].sketches['__profile__'].CircleByCenterPerimeter(center=(
    0.0, 0.0), point1=(RAD_INT, 0.0))
mdb.models['Model-1'].sketches['__profile__'].CircleByCenterPerimeter(center=(
    0.0, 0.0), point1=(RAD_EXT, 0.0))
mdb.models['Model-1'].Part(dimensionality=THREE_D, name='tube', 
    type=DEFORMABLE_BODY)
mdb.models['Model-1'].parts['tube'].BaseSolidExtrude(depth=LEN_ACT, 
    sketch=mdb.models['Model-1'].sketches['__profile__'])
del mdb.models['Model-1'].sketches['__profile__']

############################################################
# CREATE CAP					 
############################################################

mdb.models['Model-1'].ConstrainedSketch(name='__profile__', sheetSize=20.0)
mdb.models['Model-1'].sketches['__profile__'].CircleByCenterPerimeter(center=(
    0.0, 0.0), point1=(RAD_INT, 0.0))
mdb.models['Model-1'].Part(dimensionality=THREE_D, name='cap', 
    type=DEFORMABLE_BODY)
mdb.models['Model-1'].parts['cap'].BaseSolidExtrude(depth=THICK_CAP, 
    sketch=mdb.models['Model-1'].sketches['__profile__'])
del mdb.models['Model-1'].sketches['__profile__']

############################################################
# ADD INSTANCES TO ASSEMBLY					 
############################################################

mdb.models['Model-1'].rootAssembly.Instance(dependent=ON, name=
    'tube-1', part=
    mdb.models['Model-1'].parts['tube'])
mdb.models['Model-1'].rootAssembly.Instance(dependent=ON, name=
    'cap-1', part=
    mdb.models['Model-1'].parts['cap'])
mdb.models['Model-1'].rootAssembly.Instance(dependent=ON, name=
    'cap-2', part=
    mdb.models['Model-1'].parts['cap'])
mdb.models['Model-1'].rootAssembly.translate(instanceList=(
    'cap-2', ), vector=(0.0, 0.0, LEN_ACT-THICK_CAP))

############################################################
# MERGE INSTANCES 					 
############################################################

mdb.models['Model-1'].rootAssembly.InstanceFromBooleanMerge(domain=GEOMETRY, 
    instances=(
    mdb.models['Model-1'].rootAssembly.instances['tube-1'], 
    mdb.models['Model-1'].rootAssembly.instances['cap-1'],
    mdb.models['Model-1'].rootAssembly.instances['cap-2']), 
    keepIntersections=OFF, name='Merged_Actuator', originalInstances=DELETE)

del [mdb.models['Model-1'].parts['tube'],mdb.models['Model-1'].parts['cap']]

mdb.models['Model-1'].rootAssembly.Instance(dependent=ON, name=
    'Merged_Actuator-1', part=
    mdb.models['Model-1'].parts['Merged_Actuator'])

############################################################
# CREATE FIBERS
############################################################

mdb.models['Model-1'].Part(dimensionality=THREE_D, name='fibers', type=
    DEFORMABLE_BODY)
mdb.models['Model-1'].parts['fibers'].ReferencePoint(point=(0.0, 0.0, 0.0))

############################################################
# create list of points in a circle 
############################################################
a = 0
b = 200
f = 2*pi/b
angle_array = [x * f for x in range(a, b)]

############################################################
# define the fiber geometry 
############################################################
FIBER_SPACING = 2.0*pi*RAD_EXT/tan((90-ALPHA)*pi/180.0) #vertical distance between fiber loops (pitch of the helix)
LOOPS = int((LEN_ACT-2*FIBER_LEN_INIT)/FIBER_SPACING)+1 #number of fiber loops
NUMFIBERS=int(max(13.0*sin(ALPHA*pi/180.0)+0.5,1.0))
OFFSET=2.0*pi/float(NUMFIBERS)

for j in range(0,NUMFIBERS):
    point_vec=[]
    point_z=[]
    for i in range(0,len(angle_array)):
        angle=angle_array[i]
        point_vec=point_vec+[RAD_EXT*math.cos(angle+j*OFFSET),RAD_EXT*math.sin(angle+j*OFFSET),FIBER_SPACING*angle/(2.0*math.pi)]
    myPoints=()
    for i in range(0,LOOPS):
        for k in range(0,len(point_vec)/3):
            myPoints=myPoints+((point_vec[3*k],point_vec[3*k+1],point_vec[3*k+2]+i*FIBER_SPACING+FIBER_LEN_INIT),)
	    point_z=point_z+[point_vec[3*k+2]+i*FIBER_SPACING+FIBER_LEN_INIT] #cut off last loop when top of actuator is reached

    index=bisect.bisect(point_z,LEN_ACT-FIBER_LEN_INIT)
    myPoints=myPoints[0:index+1]

    mdb.models['Model-1'].parts['fibers'].WireSpline( meshable=ON, points=myPoints, smoothClosedSpline=ON)

for j in range(0,NUMFIBERS):
    point_vec=[]
    point_z=[]
    for i in range(0,len(angle_array)):
        angle=angle_array[i]
        point_vec=point_vec+[-RAD_EXT*math.cos(angle+j*OFFSET),RAD_EXT*math.sin(angle+j*OFFSET),FIBER_SPACING*angle/(2.0*math.pi)]
    myPoints=()
    for i in range(0,LOOPS):
        for k in range(0,len(point_vec)/3):
            myPoints=myPoints+((point_vec[3*k],point_vec[3*k+1],point_vec[3*k+2]+i*FIBER_SPACING+FIBER_LEN_INIT),)
	    point_z=point_z+[point_vec[3*k+2]+i*FIBER_SPACING+FIBER_LEN_INIT] #cut off last loop when top of actuator is reached

    index=bisect.bisect(point_z,LEN_ACT-FIBER_LEN_INIT)
    myPoints=myPoints[0:index+1]

    mdb.models['Model-1'].parts['fibers'].WireSpline(meshable=ON, points=myPoints, smoothClosedSpline=ON)

mdb.models['Model-1'].parts['fibers'].Set(edges=
    mdb.models['Model-1'].parts['fibers'].edges, name='FiberSet')

mdb.models['Model-1'].rootAssembly.Instance(dependent=ON, name='fibers-1', 
    part=mdb.models['Model-1'].parts['fibers'])

############################################################
# CREATE MATERIALS
############################################################

mdb.models['Model-1'].Material(name='Elastomer1')
mdb.models['Model-1'].Material(name='Elastomer2')
mdb.models['Model-1'].Material(name='Kevlar')

############################################################
# ASSIGN MATERIAL PROPERTIES					 
############################################################

mdb.models['Model-1'].materials['Elastomer1'].Hyperelastic(materialType=
    ISOTROPIC, table=(ELAST_MAT1, ), testData=OFF, type=
    YEOH, volumetricResponse=VOLUMETRIC_DATA)
mdb.models['Model-1'].materials['Elastomer2'].Hyperelastic(materialType=
    ISOTROPIC, table=(ELAST_MAT2, ), testData=OFF, type=
    YEOH, volumetricResponse=VOLUMETRIC_DATA)
mdb.models['Model-1'].materials['Kevlar'].Elastic(table=((FIBER_E, FIBER_V), ))

############################################################
# CREATE SECTIONS					 
############################################################

mdb.models['Model-1'].HomogeneousSolidSection(material='Elastomer1', name=
    'elastomer1Section', thickness=None)
mdb.models['Model-1'].HomogeneousSolidSection(material='Elastomer2', name=
    'elastomer2Section', thickness=None)
mdb.models['Model-1'].CircularProfile(name='fiberProfile', r=FIBER_R)
mdb.models['Model-1'].BeamSection(consistentMassMatrix=False, integration=
	DURING_ANALYSIS, material='Kevlar', name='kevlarSection', poissonRatio=FIBER_V, 
	profile='fiberProfile', temperatureVar=LINEAR)
mdb.models['Model-1'].parts['fibers'].SectionAssignment(offset=0.0, 
    offsetField='', offsetType=MIDDLE_SURFACE, region=
    mdb.models['Model-1'].parts['fibers'].sets['FiberSet'], sectionName=
    'kevlarSection', thicknessAssignment=FROM_SECTION)

############################################################
# PARTITION ACTUATOR
############################################################
#cap 1
idPlane=mdb.models['Model-1'].parts['Merged_Actuator'].DatumPlaneByPrincipalPlane(
    offset=THICK_CAP, principalPlane=XYPLANE).id
mdb.models['Model-1'].parts['Merged_Actuator'].PartitionCellByDatumPlane(cells=
    mdb.models['Model-1'].parts['Merged_Actuator'].cells, datumPlane=
    mdb.models['Model-1'].parts['Merged_Actuator'].datums[idPlane])
#cap 2
idPlane=mdb.models['Model-1'].parts['Merged_Actuator'].DatumPlaneByPrincipalPlane(
    offset=LEN_ACT-THICK_CAP, principalPlane=XYPLANE).id
mdb.models['Model-1'].parts['Merged_Actuator'].PartitionCellByDatumPlane(cells=
    mdb.models['Model-1'].parts['Merged_Actuator'].cells, datumPlane=
    mdb.models['Model-1'].parts['Merged_Actuator'].datums[idPlane])
#divide main tube in half
idPlane=mdb.models['Model-1'].parts['Merged_Actuator'].DatumPlaneByPrincipalPlane(
    offset=0.0, principalPlane=YZPLANE).id
mdb.models['Model-1'].parts['Merged_Actuator'].PartitionCellByDatumPlane(cells=
    mdb.models['Model-1'].parts['Merged_Actuator'].cells, datumPlane=
    mdb.models['Model-1'].parts['Merged_Actuator'].datums[idPlane])

############################################################
# ASSIGN SECTIONS					 
############################################################
#cap 1
mdb.models['Model-1'].parts['Merged_Actuator'].SectionAssignment(offset=0.0, 
    offsetField='', offsetType=MIDDLE_SURFACE, region=Region(
    cells=mdb.models['Model-1'].parts['Merged_Actuator'].cells.findAt(((
    (RAD_INT+RAD_EXT)/2.0, 0.0, 0.0), ),((-(RAD_INT+RAD_EXT)/2.0,0.0,0.0),), )), sectionName='elastomer2Section', 
    thicknessAssignment=FROM_SECTION)
#cap 2
mdb.models['Model-1'].parts['Merged_Actuator'].SectionAssignment(offset=0.0, 
    offsetField='', offsetType=MIDDLE_SURFACE, region=Region(
    cells=mdb.models['Model-1'].parts['Merged_Actuator'].cells.findAt(((
    (RAD_INT+RAD_EXT)/2.0, 0.0, LEN_ACT), ),((-(RAD_INT+RAD_EXT)/2.0,0.0,LEN_ACT),) )), sectionName='elastomer2Section', 
    thicknessAssignment=FROM_SECTION)
#main tube side 1
mdb.models['Model-1'].parts['Merged_Actuator'].SectionAssignment(offset=0.0, 
    offsetField='', offsetType=MIDDLE_SURFACE, region=Region(
    cells=mdb.models['Model-1'].parts['Merged_Actuator'].cells.findAt(((
    (RAD_EXT+RAD_INT)/2.0, 0.0, LEN_ACT/2.0), ), )), sectionName='elastomer1Section', 
    thicknessAssignment=FROM_SECTION)
#main tube side 2
mdb.models['Model-1'].parts['Merged_Actuator'].SectionAssignment(offset=0.0, 
    offsetField='', offsetType=MIDDLE_SURFACE, region=Region(
    cells=mdb.models['Model-1'].parts['Merged_Actuator'].cells.findAt(((
    -(RAD_EXT+RAD_INT)/2.0, 0.0, LEN_ACT/2.0), ), )), sectionName='elastomer2Section', 
    thicknessAssignment=FROM_SECTION)

############################################################
#CREATE SURFACES
############################################################
mdb.models['Model-1'].parts['Merged_Actuator'].Surface(name='InnerSurf', 
    side1Faces=(
    mdb.models['Model-1'].parts['Merged_Actuator'].faces.findAt(((-RAD_INT/2.0,0.0,LEN_ACT-THICK_CAP),)),
    mdb.models['Model-1'].parts['Merged_Actuator'].faces.findAt(((RAD_INT/2.0,0.0,LEN_ACT-THICK_CAP),)),
    mdb.models['Model-1'].parts['Merged_Actuator'].faces.findAt(((-RAD_INT/2.0,0.0,THICK_CAP),)),
    mdb.models['Model-1'].parts['Merged_Actuator'].faces.findAt(((RAD_INT/2.0,0.0,THICK_CAP),)),
    mdb.models['Model-1'].parts['Merged_Actuator'].faces.findAt(((-RAD_INT,0.0,LEN_ACT/2.0),)),
    mdb.models['Model-1'].parts['Merged_Actuator'].faces.findAt(((RAD_INT,0.0,LEN_ACT/2.0),)),))	

mdb.models['Model-1'].parts['Merged_Actuator'].Surface(name='OuterSurf', 
    side1Faces=(
    mdb.models['Model-1'].parts['Merged_Actuator'].faces.findAt(((RAD_EXT, 0.0,LEN_ACT/2.0), )),
    mdb.models['Model-1'].parts['Merged_Actuator'].faces.findAt(((-RAD_EXT, 0.0,LEN_ACT/2.0), )),
    mdb.models['Model-1'].parts['Merged_Actuator'].faces.findAt(((-RAD_EXT, 0.0,THICK_CAP/2.0), )),
    mdb.models['Model-1'].parts['Merged_Actuator'].faces.findAt(((-RAD_EXT, 0.0,LEN_ACT-THICK_CAP/2.0), )),))

mdb.models['Model-1'].rootAssembly.regenerate()
############################################################
# CREATE STEP
############################################################

mdb.models['Model-1'].StaticStep(name='Step-1', nlgeom=ON,
	previous='Initial',initialInc=0.05,maxNumInc=100000,minInc=0.000000001,maxInc=0.05)

############################################################
# BOUNDARY CONDITION 
############################################################

mdb.models['Model-1'].DisplacementBC(amplitude=UNSET, createStepName='Step-1', 
    distributionType=UNIFORM, fieldName='', fixed=OFF, localCsys=None, name=
    'BC-1', region=Region(
    faces=mdb.models['Model-1'].rootAssembly.instances['Merged_Actuator-1'].faces.findAt(
    ((0.0,0.0, 0.0), ), )), u1=0.0, u2=0.0, u3=0.0, ur1=UNSET, ur2=
    UNSET, ur3=UNSET)

############################################################
# APPLY LOAD
############################################################
mdb.models['Model-1'].EquallySpacedAmplitude(begin=0.0, data=(0.0, 1.0), 
	    fixedInterval=1.0, name='linear', smooth=SOLVER_DEFAULT, timeSpan=STEP)
mdb.models['Model-1'].Pressure(amplitude='linear', createStepName='Step-1', 
		    distributionType=UNIFORM, field='', magnitude=PRESSURE, name='Load-1', region=
		    mdb.models['Model-1'].rootAssembly.instances['Merged_Actuator-1'].surfaces['InnerSurf'])


############################################################
# MESH ELASTOMERS
############################################################
mdb.models['Model-1'].parts['Merged_Actuator'].seedPart(deviationFactor=0.1, 
    minSizeFactor=0.1, size=2.0)
mdb.models['Model-1'].parts['Merged_Actuator'].setMeshControls(elemShape=TET, 
    regions=
    mdb.models['Model-1'].parts['Merged_Actuator'].cells, technique=FREE)
mdb.models['Model-1'].parts['Merged_Actuator'].setElementType(elemTypes=(
    ElemType(elemCode=C3D20R, elemLibrary=STANDARD), ElemType(elemCode=C3D15, 
    elemLibrary=STANDARD), ElemType(elemCode=C3D10H, elemLibrary=STANDARD)), 
    regions=(
    mdb.models['Model-1'].parts['Merged_Actuator'].cells, ))
mdb.models['Model-1'].parts['Merged_Actuator'].generateMesh()

############################################################
# MESH FIBERS
############################################################

mdb.models['Model-1'].parts['fibers'].seedPart(deviationFactor=0.1, 
    minSizeFactor=0.1, size=1.0)
mdb.models['Model-1'].parts['fibers'].setElementType(elemTypes=(ElemType(
    elemCode=B32, elemLibrary=STANDARD), ), regions=(
    mdb.models['Model-1'].parts['fibers'].sets['FiberSet'].edges, ))
mdb.models['Model-1'].parts['fibers'].generateMesh(regions=
    mdb.models['Model-1'].parts['fibers'].sets['FiberSet'].edges)
mdb.models['Model-1'].rootAssembly.regenerate()

############################################################
# TIE CONSTRAINT
############################################################

mdb.models['Model-1'].Tie(adjust=ON, master=
    mdb.models['Model-1'].rootAssembly.instances['Merged_Actuator-1'].surfaces['OuterSurf']
    , name='Constraint-1', positionToleranceMethod=COMPUTED, slave=
    mdb.models['Model-1'].rootAssembly.instances['fibers-1'].sets['FiberSet'], 
    thickness=ON, tieRotations=ON)

############################################################
# CREATE AND RUN JOB 
############################################################

mdb.Job(atTime=None, contactPrint=OFF, description='', echoPrint=OFF, 
    explicitPrecision=SINGLE, getMemoryFromAnalysis=True, historyPrint=OFF, 
    memory=90, memoryUnits=PERCENTAGE, model='Model-1', modelPrint=OFF, 
    multiprocessingMode=DEFAULT, name=JOBNAME, nodalOutputPrecision=SINGLE, 
    numCpus=8, numDomains=8, queue=None, scratch='', type=ANALYSIS, 
    userSubroutine='', waitHours=0, waitMinutes=0)

mdb.jobs[JOBNAME].submit(consistencyChecking=OFF)
